
require "defines"
require "util"
NEConfig = {}

require "config"

game.oninit(function() OnLoad() end)
game.onload(function() OnLoad() end)

game.onevent(defines.events.onrobotbuiltentity, function(event) OnBuilt(event.createdentity) end)
game.onevent(defines.events.onbuiltentity, function(event) OnBuilt(event.createdentity) end)
game.onevent(defines.events.onpreplayermineditem, function(event) OnRemoved(event.entity) end)
game.onevent(defines.events.onrobotpremined, function(event) OnRemoved(event.entity) end)
game.onevent(defines.events.onentitydied, function(event) OnRemoved(event.entity) end)


function OnLoad()

---- Alien Control Initialization ----	
	if not glob.beacons then
      glob.beacons = {}
	end
	if not glob.minds then
      glob.minds = {}
	end
	if not glob.hiveminds then
      glob.hiveminds = {}
	end
	if not glob.minds.difficulty then
      glob.minds.difficulty = 10
	end
	if not glob.minds.difficulty2 then -- formally game.difficulty
      glob.minds.difficulty2 = 1
	end
	
---- Terraforming Initialization ----	
	if not glob.numTerraformingStations then
      glob.numTerraformingStations = 0
	end
	
---- Expansion Initialization ----	
    if not glob.Natural_Evolution_state then
		glob.Natural_Evolution_state = "Peaceful"
    end
	if not glob.Natural_Evolution_timer then
		glob.Natural_Evolution_timer = 0
	end
	if not glob.Natural_Evolution_Counter then
		glob.Natural_Evolution_Counter = 0
	end

end

---------------------------------------------
function OnBuilt(entity)
   --- Terraforming Station has been built
   if entity.name == "TerraformingStation" then
      glob.numTerraformingStations = glob.numTerraformingStations + 1
      
      glob.factormultiplier = GetFactorPerTerraformingStation(glob.numTerraformingStations)
	  writeDebug("The the number of Terraforming Stations: " .. glob.numTerraformingStations)
	  
   end

   --- Alien Control Station has been built
    if entity.name == "AlienControlStation" then
		table.insert(glob.beacons, entity)
	end
   
end

---------------------------------------------
function OnRemoved(entity)
	--- Terraforming Station has been removed
   if entity.name == "TerraformingStation" then
      glob.numTerraformingStations = glob.numTerraformingStations - 1
      
      glob.factormultiplier = GetFactorPerTerraformingStation(glob.numTerraformingStations)
   end
   
   --- Alien Control Station has been removed
	if entity.name == "AlienControlStation" then
		beaconremove()
	end
	if entity.type == "unit-spawner" and entity.force.equals(game.forces.enemy) then
		convertBase(entity, true)
	end
	
end
---- Removes the Alien Control Station ---
function beaconremove(index)
  if index then
    if glob.beacons[index] and not glob.beacons[index].valid then
      table.remove(glob.beacons, index)
      return -- if an index was found and it's entity was not valid return after removing it
    end
  end
  -- if no index was provided, or an inappropriate index was provided then loop through the table
  -- not sure if there is a better way...perhaps I could make beacons a multidimensional table
  -- and store it's index as well, then instead of searching through the table I can simply use
  -- table.remove(glob.beacons,beacon.index)...
  for k,beacon in ipairs(glob.beacons) do
    if not beacon.valid then
      table.remove(glob.beacons,k)
      writeDebug("Alien Control Station Removed")
    end
  end--not sure if there is a better way...perhaps I could make beacons a multidimensional table and store it's index as well, then instead of searching through the table I can simply use table.remove(glob.beacons,beacon.index)...
end



--------------- Terraforming Station Calculations ------------------------------

function GetFactorPerTerraformingStation(numTerraformingStations)
   local res = 1
   -- Calculate the total evolution reduction.
   for i = 1, numTerraformingStations do
      res = res + math.pow(0.9, i)
	  
   end
   -- Return the evolution reduction per Terraforming Station.
   return res / numTerraformingStations
end



---- Each time a Terraforming Station scans a sector, reduce the evolution factor ----
game.onevent(defines.events.onsectorscanned, function(event)
  if event.radar.name == "TerraformingStation" then
   game.evolutionfactor = game.evolutionfactor - ((0.00015 * glob.factormultiplier) * (1 - game.evolutionfactor))
		if game.evolutionfactor < .0001 then
			game.evolutionfactor = .0001
		end  

  writeDebug("The current Factor Multiplier is: " .. glob.factormultiplier)
   
  end
end)
--------------- END Terraforming Station ------------------------------


--------------- Alien Control Station ---------------------------------

function controlenemies()
  for k,beacon in ipairs(glob.beacons) do
    if beacon.valid then
      if beacon.energy > 0 then
        local bases = game.findentitiesfiltered{type="unit-spawner", area=getboundingbox(beacon.position, 30)} --search area of thirty around each beacon for spawners
        if #bases > 0 then
          for i, base in ipairs(bases) do
            if base.force.equals(game.forces.enemy) and math.random(glob.minds.difficulty*2)==1 then --easy = 16.5% chance, normal = 10%, hard = 5%
              convertBase(base, false)
            end
          end
        else -- no bases in range
          for i, enemy in ipairs(game.findenemyunits(beacon.position, 10)) do --search area of ten around each beacon
            if enemy.force.equals(game.forces.enemy) then --do only if not already controlled
              if math.random(glob.minds.difficulty)==1 then --easy = 33% chance, normal = 20%, hard = 10%
                enemy.force=game.player.force
                enemy.setcommand{type=defines.command.wander,distraction=defines.distraction.byenemy}
                table.insert(glob.minds, enemy)
                writeDebug("An Enemy has been Converted")
              end
            end
          end
        end
      else
        writeDebug("A Alien Control Station has no Power")
      end
    else
      beaconremove()
    end
  end
end

function removemindcontrol()
  if glob.beacons[1] then -- if there are valid beacons
    for k,mind in ipairs (glob.minds) do --remove mindcontrol from biters not in area of influence
      if not mind.valid then --first make sure the enemy is still valid, if not remove them
        table.remove(glob.minds, k)
      else -- is valid
        local controlled = false --assume out of range
        if game.findentitiesfiltered{name="AlienControlStation", area=getboundingbox(mind.position, 10)}[1] then --a AlienControlStation is in range
          controlled = true
          break
        end
        if not controlled then mind.force=game.forces.enemy end
      end
    end
  end
end

function convertBase(base, died)
  local enemies=getboundingbox(base.position, 10/glob.minds.difficulty2)
  local units={}
  local hives={}
  local worms={}
  local count=0
  enemies = game.findentities(enemies)
  for i, enemy in ipairs(enemies) do
    if enemy.type=="turret" and enemy.force.equals(game.forces.enemy) then
      table.insert(worms, enemy)
	elseif enemy.type=="small-worm-turret" and enemy.force.equals(game.forces.enemy) then
      table.insert(worms, enemy)
	elseif enemy.type=="medium-worm-turret" and enemy.force.equals(game.forces.enemy) then
      table.insert(worms, enemy)
	elseif enemy.type=="big-worm-turret" and enemy.force.equals(game.forces.enemy) then
      table.insert(worms, enemy)
    elseif enemy.type=="unit-spawner" then
      table.insert(hives, enemy)
    elseif enemy.type=="unit" then
      table.insert(units, enemy)
    end
  end
  count=#units+#hives+#worms
  if count~=0 then -- prevent empty random interval
    	
	writeDebug("The number of Units Converted: " .. count)	
   -- writeDebug(math.random(glob.minds.difficulty2 + math.pow(count, 1/4)))
  end
  if count~=0 and math.random(glob.minds.difficulty2+math.sqrt(count))==1 then
    if died then table.insert(glob.hiveminds, game.createentity{name=base.name, position=base.position, force=game.player.force}) end
    for _, worm in pairs(worms) do worm.force=game.player.force writeDebug("converted base") end
    for _, hive in pairs(hives) do hive.force=game.player.force table.insert(glob.hiveminds, hive) end
    for _, unit in pairs(units) do
      unit.force=game.player.force
      unit.setcommand{type=defines.command.wander, distraction=defines.distraction.byenemy}
      -- remove mind controlled biters in range from the minds table
      --so they aren't converted back into enemies when wandering away from the beacon
      for i, controlled in ipairs(glob.minds) do
        writeDebug{unit, controlled}
        if unit.equals(controlled) then
          table.remove(glob.minds, i)
          break
        end
      end
    end
  end
end

function getboundingbox(position, radius)
return {{position.x-radius, position.y-radius}, {position.x+radius,position.y+radius}}
end


--------------- END Alien Control Station ------------------------------



--------------- Expansion ----------------------------------------------

if NEConfig.Expansion then
	game.onevent(defines.events.ontick, function(event)

		if (game.tick % (60 * 6) == 0) and glob.beacons[1] then
			--beaconremove() --this should be unneeded with the onentitydied and onentitymined events.
			removemindcontrol() --free out of range biters
			controlenemies() --control newly in range biters
		end

		if (game.tick % (60 * 60)  == 0) and (game.evolutionfactor >= .005) and (glob.Natural_Evolution_state == "Peaceful") then
				
			local expansionChance = math.random(math.floor((game.evolutionfactor * 100) + glob.Natural_Evolution_Counter), math.floor((game.evolutionfactor * 100) + 100))

			-- For Early game, has about a 25% change to start Evolution
			if expansionChance >= 75 and expansionChance < 100 then
				Natural_Evolution_SetExpansionLevel("Awakening")
				
				
			-- Evolution Factor = 1% - 9%:
			elseif expansionChance >= 100 and expansionChance < 110 then
				Natural_Evolution_SetExpansionLevel("Phase 1")
				if game.evolutionfactor >.1 and glob.Natural_Evolution_Counter < 5 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
				
			-- Evolution Factor = 10% - 19%:
			elseif expansionChance >= 110 and expansionChance < 120 then
				Natural_Evolution_SetExpansionLevel("Phase 2")
				if game.evolutionfactor >.2 and glob.Natural_Evolution_Counter < 10 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
				
			-- Evolution Factor = 20% - 29%:
			elseif expansionChance >= 120 and expansionChance < 130 then
				Natural_Evolution_SetExpansionLevel("Phase 3")
				if game.evolutionfactor >.3 and glob.Natural_Evolution_Counter < 15 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
			
			-- Evolution Factor = 30% - 39%:
			elseif expansionChance >= 130 and expansionChance < 140 then
				Natural_Evolution_SetExpansionLevel("Phase 4")
				if game.evolutionfactor >.4 and glob.Natural_Evolution_Counter < 20 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
				
			-- Evolution Factor = 40% - 49%:
			elseif expansionChance >= 140 and expansionChance < 150 then
				Natural_Evolution_SetExpansionLevel("Phase 5")
				if game.evolutionfactor >.5 and glob.Natural_Evolution_Counter < 25 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
				
			-- Evolution Factor = 50% - 59%:
			elseif expansionChance >= 150 and expansionChance < 160 then
				Natural_Evolution_SetExpansionLevel("Phase 6")
				if game.evolutionfactor >.6 and glob.Natural_Evolution_Counter < 30 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
				
			-- Evolution Factor = 60% - 69%:
			elseif expansionChance >= 160 and expansionChance < 170 then
				Natural_Evolution_SetExpansionLevel("Phase 7")
				if game.evolutionfactor >.7 and glob.Natural_Evolution_Counter < 35 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
				
			-- Evolution Factor = 70% - 79%:
			elseif expansionChance >= 170 and expansionChance < 180 then
				Natural_Evolution_SetExpansionLevel("Phase 8")
				if game.evolutionfactor >.7 and glob.Natural_Evolution_Counter < 40 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
			
			-- Evolution Factor = 80% - 89%:
			elseif expansionChance >= 180 and expansionChance < 190 then
				Natural_Evolution_SetExpansionLevel("Phase 9")
				if game.evolutionfactor >.8 and glob.Natural_Evolution_Counter < 45 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
				
			-- Evolution Factor = 90% - 98%:
			elseif expansionChance >= 190 and expansionChance < 199 then
				Natural_Evolution_SetExpansionLevel("Phase 10")
				if game.evolutionfactor >.9 and glob.Natural_Evolution_Counter < 50 then
				glob.Natural_Evolution_Counter = glob.Natural_Evolution_Counter + 1
				end
			
			-- Evolution Factor = 99% - 100%:						
			elseif expansionChance >= 199 and expansionChance <= 200 then
				Natural_Evolution_SetExpansionLevel("Armageddon")
			end
			
		writeDebug("The Expansion Chance is: " .. expansionChance)	
		writeDebug("The Natural_Evolution Counter is: " .. glob.Natural_Evolution_Counter)	

		end
		
		if glob.Natural_Evolution_state ~= "Peaceful" then
			if glob.Natural_Evolution_timer > 0 then
				glob.Natural_Evolution_timer = glob.Natural_Evolution_timer - 1

			else
				Natural_Evolution_SetExpansionLevel("Peaceful")
			end
			
		end

	end)

	function Natural_Evolution_SetExpansionLevel(expansionState)
		expansionState = expansionState or "Peaceful"
		
		if expansionState == "Peaceful" then
			game.mapsettings.enemy_expansion.enabled = false
			glob.Natural_Evolution_timer = 0
			-- Experimental... Might be causing some issues
			-- Each time a State is triggered, the Evolution Factor is decreased slightly, just during the Phase.
			--[[
			if glob.Natural_Evolution_state == "Awakening" then
				game.evolutionfactor = game.evolutionfactor		
			elseif glob.Natural_Evolution_state == "Phase 1" then
				game.evolutionfactor = game.evolutionfactor - .00012
			elseif glob.Natural_Evolution_state == "Phase 2" then
				game.evolutionfactor = game.evolutionfactor - .00014
			elseif glob.Natural_Evolution_state == "Phase 3" then
				game.evolutionfactor = game.evolutionfactor - .00016
			elseif glob.Natural_Evolution_state == "Phase 4" then
				game.evolutionfactor = game.evolutionfactor - .00018			
			elseif glob.Natural_Evolution_state == "Phase 5" then
				game.evolutionfactor = game.evolutionfactor - .00020
			elseif glob.Natural_Evolution_state == "Phase 6" then
				game.evolutionfactor = game.evolutionfactor - .00022	
			elseif glob.Natural_Evolution_state == "Phase 7" then
				game.evolutionfactor = game.evolutionfactor - .00024
			elseif glob.Natural_Evolution_state == "Phase 8" then
				game.evolutionfactor = game.evolutionfactor - .00026
			elseif glob.Natural_Evolution_state == "Phase 9" then
				game.evolutionfactor = game.evolutionfactor - .00028
			elseif glob.Natural_Evolution_state == "Phase 10" then
				game.evolutionfactor = game.evolutionfactor - .00030
			elseif glob.Natural_Evolution_state == "Armageddon" then
				game.evolutionfactor = game.evolutionfactor - .00040
			end
			--]]
			if game.evolutionfactor < .001 then
				game.evolutionfactor = .001
			end
		
		-- Defines the values for the different Evolution States.
		elseif expansionState == "Awakening" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1 * 3600, 2 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 3
				game.mapsettings.enemy_expansion.max_expansion_distance = 5
				game.mapsettings.enemy_expansion.min_player_base_distance = 15
				game.mapsettings.enemy_expansion.settler_group_min_size = 2
				game.mapsettings.enemy_expansion.settler_group_max_size = 4
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 5 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 20 * 60

		
		elseif expansionState == "Phase 1" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(2 * 3600, 5 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 5
				game.mapsettings.enemy_expansion.max_expansion_distance = 6
				game.mapsettings.enemy_expansion.min_player_base_distance = 10
				game.mapsettings.enemy_expansion.settler_group_min_size = 4
				game.mapsettings.enemy_expansion.settler_group_max_size = 8
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 10 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 30 * 60

		elseif expansionState == "Phase 2" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.9 * 3600, 4.8 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 5
				game.mapsettings.enemy_expansion.max_expansion_distance = 8
				game.mapsettings.enemy_expansion.min_player_base_distance = 9
				game.mapsettings.enemy_expansion.settler_group_min_size = 8
				game.mapsettings.enemy_expansion.settler_group_max_size = 14
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 10 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 29 * 60

		elseif expansionState == "Phase 3" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.8 * 3600, 4.6 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 5
				game.mapsettings.enemy_expansion.max_expansion_distance = 10
				game.mapsettings.enemy_expansion.min_player_base_distance = 8
				game.mapsettings.enemy_expansion.settler_group_min_size = 12
				game.mapsettings.enemy_expansion.settler_group_max_size = 20
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 9 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 28 * 60


		elseif expansionState == "Phase 4" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.7 * 3600, 4.4 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 5
				game.mapsettings.enemy_expansion.max_expansion_distance = 12
				game.mapsettings.enemy_expansion.min_player_base_distance = 8
				game.mapsettings.enemy_expansion.settler_group_min_size = 16
				game.mapsettings.enemy_expansion.settler_group_max_size = 26
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 9 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 27 * 60

		elseif expansionState == "Phase 5" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.6 * 3600, 4.2 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 4
				game.mapsettings.enemy_expansion.max_expansion_distance = 14
				game.mapsettings.enemy_expansion.min_player_base_distance = 7
				game.mapsettings.enemy_expansion.settler_group_min_size = 20
				game.mapsettings.enemy_expansion.settler_group_max_size = 32
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 8 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 26 * 60

		elseif expansionState == "Phase 6" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.5 * 3600, 4 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 4
				game.mapsettings.enemy_expansion.max_expansion_distance = 16
				game.mapsettings.enemy_expansion.min_player_base_distance = 6
				game.mapsettings.enemy_expansion.settler_group_min_size = 24
				game.mapsettings.enemy_expansion.settler_group_max_size = 38
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 8 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 25 * 60
				
		elseif expansionState == "Phase 7" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.4 * 3600, 3.8 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 4
				game.mapsettings.enemy_expansion.max_expansion_distance = 18
				game.mapsettings.enemy_expansion.min_player_base_distance = 5
				game.mapsettings.enemy_expansion.settler_group_min_size = 28
				game.mapsettings.enemy_expansion.settler_group_max_size = 44
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 7 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 24 * 60

		elseif expansionState == "Phase 8" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.3 * 3600, 3.6 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 4
				game.mapsettings.enemy_expansion.max_expansion_distance = 20
				game.mapsettings.enemy_expansion.min_player_base_distance = 4
				game.mapsettings.enemy_expansion.settler_group_min_size = 32
				game.mapsettings.enemy_expansion.settler_group_max_size = 50
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 7 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 23 * 60


		elseif expansionState == "Phase 9" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.2 * 3600, 3.4 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 3
				game.mapsettings.enemy_expansion.max_expansion_distance = 22
				game.mapsettings.enemy_expansion.min_player_base_distance = 3
				game.mapsettings.enemy_expansion.settler_group_min_size = 36
				game.mapsettings.enemy_expansion.settler_group_max_size = 56
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 6 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 22 * 60



		elseif expansionState == "Phase 10" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1.1 * 3600, 3.2 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 3
				game.mapsettings.enemy_expansion.max_expansion_distance = 30
				game.mapsettings.enemy_expansion.min_player_base_distance = 0
				game.mapsettings.enemy_expansion.settler_group_min_size = 40
				game.mapsettings.enemy_expansion.settler_group_max_size = 100
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 5 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 20 * 60


		
		elseif expansionState == "Armageddon" then
			game.mapsettings.enemy_expansion.enabled = true
			glob.Natural_Evolution_timer = math.random(1 * 3600, 3 * 3600)
				game.mapsettings.enemy_expansion.min_base_spacing = 2
				game.mapsettings.enemy_expansion.max_expansion_distance = 50
				game.mapsettings.enemy_expansion.min_player_base_distance = 0
				game.mapsettings.enemy_expansion.settler_group_min_size = 100
				game.mapsettings.enemy_expansion.settler_group_max_size = 200
				game.mapsettings.enemy_expansion.min_expansion_cooldown = 5 * 60
				game.mapsettings.enemy_expansion.max_expansion_cooldown = 15 * 60
		
		end

				
		if expansionState ~= "Peaceful" then
		writeDebug("Expansion state set to: " .. expansionState)	
		end
		
			
	end
--------------- END Expansion ------------------------------
end

function writeDebug(message)
  if NEConfig.QCCode then game.player.print(tostring(message)) end
end
